/*
 * rcpi.c - RCPI based STA steering.
 *
 * Copyright (C) 2022 IOPSYS Software Solutions AB. All rights reserved.
 *
 * See LICENSE file for license related information.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <libubox/list.h>

#include <easy/easy.h>
#include <wifidefs.h>
#include "wifi_dataelements.h"
#include "wifi_opclass.h"

#include "utils.h"
#include "debug.h"
#include "steer.h"


#define DEFAULT_RCPI_THRESHOLD       86	/* = -67 dBm */
#define DEFAULT_RCPI_HYSTERESIS      10	/* = 5dBm */
#define DEFAULT_RCPI_DIFFSNR         16	/* = 8dB */

struct rcpi_steer_control {
	uint8_t diffsnr;
	uint8_t low;
	uint8_t hysteresis;
	uint8_t rcpi_threshold;
	uint8_t max_btm_attempt;
	bool bandsteer;
	void *self;
};

static int rcpi_steer_init(void **priv);
static int rcpi_steer_config(void *priv, struct steer_config *cfg);
static int rcpi_steer_exit(void *priv);

static bool is_bandsteer_allowed(void *priv)
{
	struct rcpi_steer_control *sctrl = (struct rcpi_steer_control *)priv;

	if (sctrl->bandsteer) {
		cntlr_dbg(LOG_STEER,
			  "%s: bandsteer is allowed\n",
			  __func__);
		return true;
	}

	cntlr_dbg(LOG_STEER,
		  "%s: bandsteer is not allowed\n",
		  __func__);
	return false;
}

int rcpi_steer(void *priv, struct steer_sta *s)
{
	cntlr_trace(LOG_STEER, "%s: --------->\n", __func__);

	struct rcpi_steer_control *sctrl = (struct rcpi_steer_control *)priv;
	struct wifi_sta_meas_report *b = NULL;
	uint8_t current_bss_rcpi = s->sta->rcpi;

	s->verdict = STEER_VERDICT_UNDECIDED;
	s->reason = STEER_REASON_UNDEFINED;
	s->mode = STEER_MODE_UNDEFINED;
	cntlr_dbg(LOG_STEER,
		  "%s: STA " MACFMT " has RCPI = %u\n",
		  __func__, MAC2STR(s->sta->macaddr),
		  current_bss_rcpi);

	cntlr_dbg(LOG_STEER,
		  "%s: Configured RCPI steering threshold = %u\n",
		  __func__, sctrl->rcpi_threshold);

	if (current_bss_rcpi >= sctrl->rcpi_threshold) {
		/* RCPI stil above threshold level */
		cntlr_dbg(LOG_STEER,
			  "%s: STA " MACFMT " has RCPI above steering threshold\n",
			  __func__, MAC2STR(s->sta->macaddr));

		s->verdict = STEER_VERDICT_NOK; /* TODO: STEER_VERDICT_MAYBE */
		return 0;
	}

	if (s->sta->mapsta.failed_steer_attempts > sctrl->max_btm_attempt) {
		/* skip further checks for sticky candidates */
		cntlr_dbg(LOG_STEER,
			  "%s: STA " MACFMT " has been steered %d times."
			  " Mark it sticky and skip\n",
			  __func__, MAC2STR(s->sta->macaddr),
			  s->sta->mapsta.failed_steer_attempts);

		s->verdict = STEER_VERDICT_EXCLUDE;
		return 0;
	}

	if (list_empty(s->meas_reportlist)) {
		cntlr_dbg(LOG_STEER,
			  "%s: No Beacon reports available yet from STA " MACFMT "\n",
			  __func__, MAC2STR(s->sta->macaddr));
		s->verdict = STEER_VERDICT_NOK;
		return 0;
	}

	/* Set temporary best - fresh for current BSS */
	s->best = NULL;
	list_for_each_entry(b, s->meas_reportlist, list) {
		if (!memcmp(b->bssid, s->bssid, 6))
			cntlr_dbg(LOG_STEER,
				  "%s: local bssid ("MACFMT") measurement found stale:%d rcpi:%u\n",
				  __func__, MAC2STR(s->bssid), b->stale, b->rcpi);
		if (!b->stale && !memcmp(b->bssid, s->bssid, 6) && !s->best) {
			/* Current BSS's measurement is set as tmp best */
			current_bss_rcpi = b->rcpi;
			s->best = b;
			cntlr_dbg(LOG_STEER,
				  "%s: new best (local) rcpi:%u based on measurement report\n",
				  __func__, current_bss_rcpi);
			//break;
		}
	}

	if (s->best == NULL) {
		cntlr_dbg(LOG_STEER,
			  "%s: local bssid was not found with active measurement in list, search for first non-stale measurement\n", __func__);
		/* Missing fresh meas for current BSS - use first non stale entry */
		list_for_each_entry(b, s->meas_reportlist, list) {
			if (b->stale)
				continue;

			if (!is_bandsteer_allowed(priv)
			    /* Consider tBSS only from STA's current band */
			    && wifi_opclass_get_band(b->opclass) != s->band)
				continue;
			s->best = b;
			break;
		}
	}

	if (s->best == NULL) {
		/* Only stale measurements on the list, do not steer */
		cntlr_dbg(LOG_STEER, "%s: missing fresh reports\n", __func__);
		s->verdict = STEER_VERDICT_NOK;
		return 0;
	}

	/* Find measurement with best RCPI on the list */
	list_for_each_entry(b, s->meas_reportlist, list) {
		if (b->rcpi > s->best->rcpi) {

			/* Use only non-stale measurements */
			if (b->stale) {
				cntlr_dbg(LOG_STEER, "%s: ignoring stale bcn report\n", __func__);
				continue;
			}

			cntlr_err(LOG_STEER, "%s: beacon-resp: {  bssid:"MACFMT", band:%d, opclass:%u, channel:%u }, station: { band: %d }\n",
				  __func__, MAC2STR(b->bssid), wifi_opclass_get_band(b->opclass), b->opclass, b->channel, s->band);

			if (!is_bandsteer_allowed(priv)
			    /* Consider tBSS only from STA's current band */
			    && wifi_opclass_get_band(b->opclass) != s->band)
				continue;

			cntlr_dbg(LOG_STEER, "%s: new best bcn from "MACFMT" with rcpi %d\n",
				  __func__, MAC2STR(b->bssid), b->rcpi);

			s->best = b;
		}
	}

	if (s->best->rcpi - current_bss_rcpi < sctrl->diffsnr) {
		/* New best rcpi is not good enough */
		cntlr_dbg(LOG_STEER, "%s: best_rcpi[%d] - curr_rcpi[%d] < diff_rcpi[%d]\n",
			  __func__, s->best->rcpi, current_bss_rcpi, sctrl->diffsnr);
		s->verdict = STEER_VERDICT_NOK;
		return 0;
	}

	/* Best bssid is a current one - no need to steer */
	if (!memcmp(s->best->bssid, s->bssid, 6)) {
		cntlr_dbg(LOG_STEER, "%s: own BSS is best - no need to steer\n", __func__);
		s->verdict = STEER_VERDICT_NOK;
		return 0;
	}

	s->reason = STEER_REASON_LOW_RCPI;
	s->verdict = STEER_VERDICT_OK;
	if (s->sta->mapsta.failed_steer_attempts < sctrl->max_btm_attempt)
		s->mode = STEER_MODE_BTM_REQ;
	else if (s->sta->mapsta.failed_steer_attempts == sctrl->max_btm_attempt)
		s->mode = STEER_MODE_ASSOC_CTL;

	cntlr_dbg(LOG_STEER,
		  "%s: decision to steer STA " MACFMT " to new AP " MACFMT " with RCPI = %u\n",
		  __func__, MAC2STR(s->sta->macaddr), MAC2STR(s->best->bssid), s->best->rcpi);

	return 0;
}

extern struct steer_control rcpi;
struct steer_control rcpi = {
	.name = "rcpi",
	.init = rcpi_steer_init,
	.config = rcpi_steer_config,
	.exit = rcpi_steer_exit,
	.steer = rcpi_steer,
};

static int rcpi_steer_init(void **priv)
{
	struct rcpi_steer_control *p;

	p = calloc(1, sizeof(struct rcpi_steer_control));
	if (!p)
		return -1;

	*priv = p;
	p->self = &rcpi;
	p->low = 90;
	/* Initial values to be updated with steer_config */
	p->rcpi_threshold = DEFAULT_RCPI_THRESHOLD;
	p->hysteresis = DEFAULT_RCPI_HYSTERESIS;
	p->diffsnr = DEFAULT_RCPI_DIFFSNR;
	p->max_btm_attempt = DEFAULT_MAX_BTM_ATTEMPT;

	cntlr_info(LOG_STEER, "%s", "RCPI based STA steering initalized");
	return 0;
}

static int rcpi_steer_config(void *priv, struct steer_config *cfg)
{
	struct rcpi_steer_control *p = (struct rcpi_steer_control *)priv;

	if (!p)
		return -1;

	p->rcpi_threshold = cfg->rcpi_threshold;
	p->hysteresis = cfg->rcpi_hysteresis;
	p->diffsnr = cfg->rcpi_diffsnr;
	p->bandsteer = cfg->bandsteer;
	p->max_btm_attempt = cfg->max_btm_attempt;

	cntlr_dbg(LOG_STEER, "%s\n", __func__);

	return 0;
}

static int rcpi_steer_exit(void *priv)
{
	struct rcpi_steer_control *p = (struct rcpi_steer_control *)priv;

	if (p)
		free(p);

	cntlr_dbg(LOG_STEER, "%s\n", __func__);
	return 0;
}

